"""Interface and Implementation for generating plots and charts.

The default implementation of :py:class:`.AbstractPlot` is the
:py:class:`.Plot` class which internally uses the matplotlib library
to generate the plots. If you want to use another library to generate
your plots you can add a new `CustomPlot` class which implements the
:py:class:`.AbstractPlot` interface. Now your plots need to inherit
from your `CustomPlot` class and you are good to go.
"""
from __future__ import unicode_literals
from six import add_metaclass
from abc import abstractmethod, ABCMeta
import matplotlib.pyplot as plt


@add_metaclass(ABCMeta)
class AbstractPlot(object):

    """Abstract Interface for a plot class.

    Arguments:
        repo (AbstractRepository): Repository Class to fetch data.
        config (Config): Config data for the plot.
        plot_type (str, optional): Defaults to 'plot'. Type of plot
        plot (object): Plotting library
    """

    def __init__(self, repo, config, plot_type='plot'):
        """Initialize abstract plot.

        Args:
            repo (AbstractRepository): Repository Class to fetch data.
            config (Config): Config data for the plot.
            plot_type (str, optional): Defaults to 'plot'. Type of plot
        """
        self.repo = repo
        self.config = config
        self.type = plot_type
        self.plot = None

    def data(self, experiments):
        """Return Data for the plot.

        Args:
            experiments (AbstractRepository): Repository with current experiment

        Returns:
            list|dict: x,y,z values or list of x,y,z coordinates (e.g. {'x': 0, 'y': 0, 'z': 0})
        """
        return {'x': 0, 'y': 0, 'z': 0}

    @abstractmethod
    def draw(self, data, plot_idx=None):
        """Draw the plot data.

        Args:
            data (dict): Dictionary with coordinates (e.g. {'x': 0, 'y': 0, 'z': 0})
            plot_idx (integer, optional): Defaults to None. Index for when multiple plots are drawn

        Raises:
            NotImplementedError: Must implement draw method
        """
        raise NotImplementedError('Must implement draw method!')

    @abstractmethod
    def labeling(self):
        """Add labels to the plot.

        Raises:
            NotImplementedError: Must implement labeling method
        """
        raise NotImplementedError('Must implement labeling method!')

    @abstractmethod
    def plotting(self):
        """Generate the plot, i.e. add labels, titles, legend etc and draw the plot.

        Returns:
            object: Plot object

        Raises:
            NotImplementedError: Must implement plotting method
        """
        raise NotImplementedError('Must implement plotting method!')

    @abstractmethod
    def save(self, filename):
        """Save the plot.

        Args:
            filename (str): Filename to save the plot.

        Raises:
            NotImplementedError: Must implement save method
        """
        raise NotImplementedError('Must implement save method!')

    @abstractmethod
    def show(self):
        """Show the plot.

        Raises:
            NotImplementedError: Must implement show method
        """
        raise NotImplementedError('Must implement show method!')


class Plot(AbstractPlot):

    """Concrete Implementation of the AbstractPlot Interface for a plot class.

    This implementation uses matplotlib to genereate the plots.

    Arguments:
        repo (AbstractRepository): Repository Class to fetch data.
        config (Config): Config data for the plot.
        plot_type (str, optional): Defaults to 'plot'. Type of plot
        plot (matplotlib.pyplot): matplotlib.pyplot module
    """

    def __init__(self, repo, config, plot_type='plot'):
        """Initialize abstract plot.

        Args:
            repo (AbstractRepository): Repository Class to fetch data.
            config (Config): Config data for the plot.
            plot_type (str, optional): Defaults to 'plot'. Type of plot
        """
        super(Plot, self).__init__(repo, config, plot_type)
        self.plot = plt

    def labeling(self):
        """Add labels to the plot."""
        if self.config.get('labels'):
            self.plot.xlabel(self.config.get('labels.x-axis'))
            self.plot.ylabel(self.config.get('labels.y-axis'))

        if self.config.get('title'):
            self.plot.title(self.config.get('title.label'), loc=self.config.get('title.loc'))

        if self.config.get('legend'):
            self.plot.legend(**self.config.get('legend'))

    def plotting(self):
        """Generate the plot, i.e. add labels, titles, legend etc and draw the plot.

        Returns:
            matplotlib.pyplot: Plot object
        """
        self.plot.figure()
        exps = self.repo.get(['name', self.config.get('experiment')])
        plot_data = self.data(exps)

        if isinstance(plot_data, list):
            for plot, data in enumerate(plot_data):
                self.draw(data, plot)
        else:
            self.draw(plot_data)

        if self.config.get('styles.grid', False):
            self.plot.grid(True)

        if self.config.get('styles.axis', False):
            self.plot.axis(self.config.get('styles.axis'))

        if self.config.get('styles.ticks', False):
            self.plot.xticks(**self.config.get('styles.ticks.xticks', {}))
            self.plot.yticks(**self.config.get('styles.ticks.yticks', {}))

        self.labeling()

        return self.plot

    def draw(self, data, plot_idx=None):
        """Draw the plot data.

        Args:
            data (dict): Dictionary with coordinates (e.g. {'x': 0, 'y': 0, 'z': 0})
            plot_idx (integer, optional): Defaults to None. Index for when multiple plots are drawn
        """
        if self.type == 'histogram':
            self.plot.hist(data.get('y'), **self._get_params(plot_idx))
        elif self.type == 'errorbar':
            self.plot.errorbar(data.get('x'), data.get('y'), **self._get_params(plot_idx))
        elif self.type == 'bar':
            self.plot.bar(data.get('x'), data.get('y'), **self._get_params(plot_idx))
        elif self.type == 'pie':
            self.plot.pie(data.get('x'), **self._get_params(plot_idx, ['label']))
        elif self.type == 'scatter':
            self.plot.scatter(data.get('x'), data.get('y'), **self._get_params(plot_idx))
        elif self.type == 'polar':
            self.plot.polar(data.get('theta'), data.get('r'))
        else:
            params = self._get_params(plot_idx)
            fmt = self.config.get('styles.fmt', [])

            if plot_idx is not None and isinstance(fmt, list) and plot_idx < len(fmt):
                fmt = fmt[plot_idx]

            self.plot.plot(data.get('x'), data.get('y'), fmt, **params)

    def save(self, filename):
        """Save the plot.

        Args:
            filename (str): Filename to save the plot.
        """
        self.plot.savefig(filename)

    def show(self):
        """Show the plot."""
        self.plot.show()

    def _get_params(self, plot_idx, remove=None):
        """Get params for the plot.

        Args:
            plot_idx (integer): Index for when multiple plots are drawn.
            remove (list, optional): Defaults to None. List of keys to remove from params.

        Returns:
            dict: Dictionary of params
        """
        _params = {}
        remove = remove if remove is not None else []
        params = self.config.get('params', {})
        params['label'] = self.config.get('styles.label', '')

        for key in remove:
            del params[key]

        for key, values in params.items():
            if plot_idx is not None and isinstance(values, list) and len(values) > plot_idx:
                _params[key] = values[plot_idx]
            else:
                _params[key] = values

        return _params



# -*- coding: utf-8 -*-
"""Contains all views/routes for the migrations.

Available Routes for:
    * Upgrading Migrations
    * Downgrading Migrations
    * Refreshing Migrations
    * Getting Migration Status
"""
from flask import Blueprint, current_app, jsonify, render_template, request
from experimentum.WebGUI.helpers import ansi_escape, capture_print


blueprint = Blueprint('migrations', __name__)


def get_migration_status():
    """Get the current migration status.

    Returns:
        list: list of migrations with their status (i.e. migrated or not)
    """
    status = current_app.config.get('container').make('migrator').status(printing=False)
    status = list(map(lambda s: [ansi_escape(s[0]), ansi_escape(s[1]).lower()], status))

    return status


@blueprint.route('/upgrade')
def upgrade():
    """Upgrade the migration and send a json status message.

    Returns:
        Response: json status message
    """
    response = None
    try:
        with capture_print(escape=True) as content:
            current_app.config.get('container').make('migrator').up()

        response = {'message': content.get_text(), 'status': 'success'}
    except SystemExit:
        response = {'message': content.get_text(), 'status': 'error'}

    return jsonify(response)


@blueprint.route('/downgrade')
def downgrade():
    """Downgrade the migration and send a json status message.

    Returns:
        Response: json status message
    """
    response = None
    try:
        with capture_print(escape=True) as content:
            current_app.config.get('container').make('migrator').down()

        response = {'message': content.get_text(), 'status': 'success'}
    except SystemExit:
        response = {'message': content.get_text(), 'status': 'error'}

    if '×' in response['message']:
        response['status'] = 'error'

    return jsonify(response)


@blueprint.route('/refresh')
def refresh():
    """Refresh all migrations and send a json status message.

    Returns:
        Response: json status message
    """
    current_app.config.get('container').make('migrator').refresh()
    return jsonify({'message': '› Refreshed all migrations', 'status': 'success'})


@blueprint.route('/make', methods=['POST'])
def make():
    """Make a new migration.

    Returns:
        Response: json status message
    """
    response = None
    try:
        with capture_print(escape=True) as content:
            current_app.config.get('container').make('migrator').make(request.form['name'])

        response = {'message': content.get_text(), 'status': 'success'}
    except SystemExit:
        response = {'message': content.get_text(), 'status': 'error'}

    return jsonify(response)


@blueprint.route('/status')
def status():
    """Return the migration status as an html template.

    Returns:
        Response: HTML migration status template
    """
    return jsonify(render_template('dashboard/migrations.jinja', migrations=get_migration_status()))



"""Experiment CLI commands to allow you to run and manage your experiments.

Running experiments
-------------------
Use the ``experiments:run`` command to run an experiment.

Arguments:

====  ======================
name  Name of the experiment
====  ======================

Options:

--config=file       Use alternative config file *(relative to experiments folder)*.
--progress          Toggle visibility of the progress bar.
--n=number          Run the experiment *n* times.
--hide_performance  Hides the performance table.
-h, --help          Show the help message.

Listing experiments
-------------------
Use the ``experiments:list`` command to list status information about all experiments.

Options:

-h, --help  Show the help message.
"""
from tabulate import tabulate
from termcolor import colored
from experimentum.cli import print_failure
from experimentum.Commands import command
from experimentum.Experiments import Experiment


@command('Load an run an experiment', help='Run an experiment', arguments={
    'name': {
        'type': str, 'help': 'Name of the experiment.'
    },
    '--n': {
        'type': int, 'default': 100, 'help': 'Run the experiment *n* times.'
    },
    '--config': {
        'type': str, 'help': 'Use alternative config file relative to experiments folder.'
    },
    '--progress': {
        'action': 'store_true', 'help': 'Toggle visibility of the progress bar'
    },
    '--hide_performance': {
        'action': 'store_true', 'help': 'Hides the performance table.'
    }
})
def run(app, args):
    """Load an run an experiment.

    Args:
        app (App): App Service Container.
        args (argparse.Namespace): Command Arguments and Options.
    """
    experiment = app.make('experiment', args.name)

    if args.config is not None:
        experiment.config_file = args.config

    if args.progress is True:
        experiment.show_progress = True

    if args.hide_performance is True:
        experiment.hide_performance = True

    experiment.start(args.n)


@command('Gather status informations about all available experiments', help='List experiments')
def status(app):
    """List experiment status.

    Args:
        app (App): App Service Container.
    """
    data = []
    headers = [
        colored('Experiment', 'yellow'),
        colored('Config File', 'yellow'),
        colored('Times Executed', 'yellow')
    ]

    try:
        exps = Experiment.get_status(app)
        for experiment in exps.values():
            data.append([
                colored(experiment['name'], 'cyan'),
                colored(experiment.get('config_file'), 'cyan'),
                colored(experiment['count'], 'cyan')
            ])
    except Exception as exc:
        print_failure(exc, 2)

    print(tabulate(data, headers=headers, tablefmt='psql'))



"""Handles database migrations."""
import inflection


class Migration(object):

    """Base Migration Class."""
    revision = None

    def __init__(self, app):
        """Init the Migration and set up the schema class.

        Arguments:
            app (App): App class
        """
        self.schema = app.make('schema')

    def up(self):
        """Revert the migrations."""
        pass

    def down(self):
        """Revert the migrations."""
        pass

    def __repr__(self):
        """Return Migration name."""
        return '{}_{}'.format(
            self.revision,
            inflection.underscore(self.__class__.__name__)
        )



"""Easily add CLI commands to the app.

There are to possible ways for defining CLI commands in
order for them to be added to the :py:mod:`.CommandManager`.

Decorator
---------
The easiest method to define a command is via the :py:func:`.AbstractCommand.command`
decorator. The decorator accepts some arguments like description or help text in order
to descrine the command while the decorated function handles the command execution.
The ``arguments`` argument accepts a dictionary which each key is the name of the argument
and each value is passed to the :py:func:`argparse.add_argument` method::

    @command(
        'Some description about what the command does.',
        arguments={
            'integers': {'help': 'Some Help', 'default': 42, 'nargs': '?'},
            '--bar': {'action': 'store_true', 'help': 'Bar Help'}
        }
        help='Short help text.'
    )
    def foo(app, args):
        print(args)


Class-based
-----------
The other way of defining a command is a class based approach. Your command class
has to derived from the :py:class:`.AbstractCommand` class. Just like the decorator
you can define description, arguments and a help text. The ``handle`` method handles
the command execution::

    class FooCommand(AbstractCommand):
        description = 'Some description about what the command does'
        arguments = {
            'integers': {'help': 'Some Help', 'default': 42, 'nargs': '?'},
            '--bar': {'action': 'store_true', 'help': 'Bar Help'}
        }
        help='Short help text.'

        def handle(self, app, args):
            print(args)
"""


def command(description='', arguments=None, help=''):
    """Command decorator, creates a Command to use with the CommandManager.

    Args:
        description (str, optional): Defaults to ''. Description of the command.
        arguments (dict, optional): Defaults to None. Arguments for the command.
        help (str, optional): Defaults to ''. Help text of the command.

    Returns:
        function:
    """
    def command_decorator(command):
        """Decorator to specific a CLI command-

        Args:
            command (function): Command Handler

        Returns:
            function: Command Wrapper
        """
        def command_wrapper(args=None):
            """Wrapper to wrap handler in AbstractCommand Class

            Args:
                args (dict, optional): Defaults to None. Command Arguments

            Returns:
                AbstractCommand: Command Wrapper
            """
            cmd = AbstractCommand()
            cmd.setup(description, arguments, help)
            cmd.handle = command
            return cmd
        return command_wrapper
    return command_decorator


class AbstractCommand(object):

    """Abstract Command Class.

    Attributes:
        description (str): Description of the command.
        arguments (dict): Optional aguments for the command.
        help (str): Help Text for the command.
        args (dict): Dictionary with possible passed arguments.
    """
    description = ''
    help = ''
    arguments = {}
    args = {}

    def setup(self, description='', arguments=None, help=''):
        """Set up the command.

        Args:
            description (str, optional): Defaults to ''. Description of the command.
            arguments (dict, optional): Defaults to None. Optional aguments for the command.
            help (str, optional): Defaults to ''. Help Text for the command.
        """
        self.description = description
        self.arguments = arguments if arguments is not None else {}
        self.help = help

    def handle(self, app, args):
        """Handle the command execution.

        Args:
            app (App): Main App class
            args (dict): Dictionary with possible passed arguments.

        Raises:
            NotImplementedError: must be implemented
        """
        raise NotImplementedError(
            'handle-Method has not been implemented yet.'
        )



# -*- coding: utf-8 -*-
"""Contains all views/routes for the experiments.

It should call ``experiments:run`` to run an experiment,
show the ouput log and performance table, and show the
generated plots and charts if available.

Available Routes for:
    * Running Experiments
    * Showing Plots and Charts of experiments
"""
from flask import Blueprint, Response, request, render_template, current_app, abort
from experimentum.WebGUI.helpers import CapturedContent
from threading import Thread
import json
import os

blueprint = Blueprint('experiments', __name__)


def run_experiment(exp_thread, capturer, experiment):
    """Experiment runner, which monitors and yields the thread output.

    Args:
        exp_thread (Thread): Thread which runs the experiment.
        capturer (CapturedContent): Capturer the capture the thread output.
        experiment (Experiment): experiment measurement.
    """
    from time import sleep
    from experimentum.WebGUI.helpers import ansi_escape

    yield 'data: {}\n\n'.format(json.dumps({'type': 'started'}))

    while exp_thread.isAlive():
        sleep(.1)  # artifical delay, otherwise loop runs to fast and misses some output :/
        error = capturer.has_error()
        content = capturer.get_text()

        # if there is data in the output stream, yield it to the event stream
        if content:
            data = {'data': content, 'error': error, 'type': 'log'}
            yield 'data: {}\n\n'.format(json.dumps(data))
            capturer.clear()

    # get performance table
    points = experiment.performance.export(metrics=True)
    table = ansi_escape(experiment.performance.formatter.get_table(points, 'html'))
    yield 'data: {}\n\n'.format(json.dumps({'table': table, 'type': 'table'}))

    # Revert streams back to normal and finish event stream.
    capturer.revert()
    data = {
        'start': experiment.repos['experiment'].start.isoformat(),
        'finished': experiment.repos['experiment'].finished.isoformat(),
        'config_file': experiment.repos['experiment'].config_file,
        'config_content': experiment.repos['experiment'].config_content
    }
    yield 'data: {}\n\n'.format(json.dumps({'data': data, 'type': 'finished'}))


@blueprint.route('/run/<experiment>', methods=['GET', 'POST'])
def run(experiment):
    """Run an experiment.

    Args:
        experiment (str): Name of the experiment

    Returns:
        str|Response: HTML template, or event stream for experiment log
    """
    # Try to load the experiment
    try:
        exp = current_app.config.get('container').make('experiment', experiment)
    except SystemExit:
        abort(404)

    # Show results page for experiment
    if request.method == 'POST':
        context = {
            'iterations': request.form['iterations'],
            'config': request.form['config'],
            'experiment': experiment
        }
        return render_template('experiments/result.jinja', **context)

    # Run experiment and stream output log
    elif request.headers.get('accept') == 'text/event-stream':
        # Use submitted config and iteration
        exp.show_progress = True
        exp.config_file = request.args.get('config')
        iterations = int(request.args.get('iterations', 100))

        # Run the experiment in a seperate thread, so we can monitor its output
        capturer = CapturedContent(True)
        exp_thread = Thread(target=lambda: exp.start(iterations))
        exp_thread.deamon = True
        exp_thread.start()

        return Response(
            run_experiment(exp_thread, capturer, exp),
            content_type='text/event-stream'
        )

    # Show experiment form
    return render_template('experiments/form.jinja', name=experiment, config=exp.config_file or '')


@blueprint.route('/plots/<experiment>')
def plots(experiment):
    """Show plots and charts for experiment.

    Args:
        experiment (str): Name of experiment

    Returns:
        str: HTML template
    """
    from werkzeug.utils import secure_filename

    path = os.path.join(current_app.config['UPLOAD_FOLDER'], experiment)
    if not os.path.exists(path):
        os.makedirs(path)

    plots = [f for f in os.listdir(path) if not f.startswith('.')]
    data = {
        'experiment': experiment,
        'plots': map(secure_filename, plots)
    }

    return render_template('experiments/plots.jinja', **data)


# @blueprint.route('/streaming')
# def stream():
#     from flask import current_app, stream_with_context
#     def stream_template(template_name, **context):
#         current_app.update_template_context(context)
#         t = current_app.jinja_env.get_template(template_name)
#         rv = t.stream(context)
#         # uncomment if you don't need immediate reaction
#         # rv.enable_buffering(5)
#         return rv

#     def generate():
#         from time import sleep
#         sleep(1.5)
#         for idx, item in enumerate(data):
#             yield {'data': item, 'error': idx == 3}
#             sleep(.3)  # artifical delay

#     return Response(stream_with_context(stream_template('streaming.html', rows=generate())))



# -*- coding: utf-8 -*-
"""Dispatch and configure the flask web app.

Starting the Server
-------------------
Use the ```webgui`` command to dispatch a new server instance
(http://flask.pocoo.org/docs/1.0/patterns/appdispatch/#app-dispatch).


Options:

--debug      Enable the Debugger.
--port       Specifiy the port to run server on.
--no-reload  Disable the reload when file changes are detected.
-h, --help   Show the help message.
"""
from experimentum.Commands import command


@command(
    'Start a new web server to act as a GUI on localhost:5000.',
    arguments={
        '--port': {
            'help': 'Port number to start the server on', 'default': 5000, 'type': int
        },
        '--no-reload': {
            'help': 'Disable reload when file changes are detected',
            'dest': 'reload',
            'action': 'store_false'
        },
        '--debug': {
            'help': 'Enable the debugger', 'dest': 'debug', 'action': 'store_true'
        }
    },
    help='Starts a web server on localhost'
)
def start(app, args):
    """Start a simple server and run the flask app.

    Args:
        app (App): Main Service Container
        args (argparse.Namespace): Command Arguments and Options.
    """
    reloader = False if hasattr(args, 'reload') and args.reload is False else True
    debugger = True if hasattr(args, 'debug') and args.debug is True else False

    server = app.make('server')
    server.run(args.port, debugger, reloader)



"""Utility functions to work with files and loading classes."""
from experimentum.cli import print_failure
import glob
import os
import imp
import re


def find_files(root, path, search=None, remove='.py'):
    """Find files in a folder (except _*.py files).

    Can also search for specific filenames (case ignored).

    Args:
        root (str): Root Path
        path (str): Path to folder
        search (str, optional): Defaults to None. Filename to search for
        remove (str, optional): Defaults to '.py'. Exclude from filename comparison

    Returns:
        list: List of filenames
    """
    regex = re.compile(remove, re.IGNORECASE)
    files = glob.glob(os.path.join(root, path, '[!_]*.py'))

    if search:
        files = list(filter(
            lambda file: search.lower() == re.sub(regex, '', os.path.basename(file)).lower(),
            files
        ))

    return files


def get_basenames(root, path, remove='.py'):
    """Get file basenames of a folder.

    Args:
        root (str): Root path
        path (str): Path to folder
        remove (str, optional): Defaults to '.py'. Part to remove from filename.

    Returns:
        list: list of names
    """
    regex = re.compile(remove, re.IGNORECASE)
    files = find_files(root, path, remove=remove)

    return list(map(
        lambda file: re.sub(regex, '', os.path.basename(file)),
        files
    ))


def load_class(src, module, subclass=None):
    """Try to load a class from a module.

    Args:
        src (str): Source file
        module (str): Module name
        subclass (object, optional): Defaults to None. Object the class needs to be derived from

    Returns:
        object: loaded class
    """
    with open(src, 'rb') as filehandler:
        # try to load class from module
        try:
            mod = imp.load_source(module, src, filehandler)
            klass = getattr(mod, os.path.basename(src)[:-3])
        except Exception as exc:
            print_failure('Could not load file: %s' % str(exc), exit_code=2)

        # Check if class is derived from subclass
        if subclass and issubclass(klass, subclass) is False:
            msg = '{} must be derived from the {}.{} class'.format(
                klass.__name__,
                subclass.__module__,
                subclass.__name__
            )
            print_failure(msg, exit_code=3)

        return klass



# -*- coding: utf-8 -*-
"""Generate needed files and folders to get up and running.

In order to create all needed files and folders *(including config and migrations)*
the experimentum framework contains a quickstart command. Just run ``experimentum-quickstart``
and answer the questions.

You can also add a different root path with the ``--root`` option, e.g.::

    experimentum-quickstart --root example
"""
from __future__ import print_function
import sys
import os
import inflection
import datetime
from time import sleep
from termcolor import colored
from experimentum import __version__
from experimentum.cli import get_input
from experimentum.files import create_config_file, create_folder, create_from_stub


# Some default constants
APP = 'experimentum'
VERSION = __version__.__version__
ROOT = '.'


def _create_migration(path, name, upgrade, down):
    """Create a migration file.

    Args:
        path (str): Path to migrations folder.
        name (str): Name of Migration
        upgrade (str): Up method content
        down (str): down method content
    """
    revision = datetime.datetime.now().strftime('%Y%m%d%H%M%S')
    filename = "{}/{}_{}.py".format(os.path.join(ROOT, path), revision, name.lower())
    attrs = {
        'migration': inflection.camelize(name),
        'name': name,
        'revision': revision,
        'up': upgrade,
        'down': down
    }
    create_from_stub('Migration.stub', filename, attrs)


def _create_repository(path, name, table, attributes, nullable=None, relationships=None):
    # Relationships and Imports
    relationships = relationships if relationships is not None else {}
    nullable = nullable if nullable is not None else []
    imports = []
    relations_data = []
    for attr, repo in relationships.items():
        relations_data.append("        '{}': [{}]".format(attr, repo))
        imports.append('from {0}.{1} import {1}'.format(os.path.basename(path), repo))

    if len(relations_data):
        relationships = '{\n%s\n    }' % '\n'.join(relations_data)
        imports = '\n' + '\n'.join(imports)
    else:
        relationships = '{}'
        imports = ''

    # other attrs
    filename = "{}/{}.py".format(os.path.join(ROOT, path), name)
    attrs = {
        'name': name,
        'table': table,
        'imports': imports,
        'relationships': relationships,
        'kwargs': ', '.join(
            map(lambda attr: '{}{}'.format(attr, '=None' if attr in nullable else ''), attributes)
        ),
        'set_attr': '\n        '.join(map('self.{0} = {0}'.format, attributes))
    }

    # print(attrs)
    create_from_stub('Repository.stub', filename, attrs)


def main():
    """Generate needed files and folders to get up and running."""
    # Set Root
    global ROOT
    if len(sys.argv) == 3 and sys.argv[1] == '--root':
        ROOT = os.path.join('.', sys.argv[2])

    # Welcome Text
    print(colored('Welcome to the {} {} quickstart utility.'.format(APP, VERSION), attrs=['bold']))
    print('\nPlease enter values for the following settings (just press Enter to')
    print('accept a default value, if one is given in brackets)\n')
    print(colored('Selected root path: {}\n'.format(ROOT), attrs=['bold']))

    # Get Config from User
    _folders = ['config', 'migrations', 'repositories', 'experiments', 'logs']
    folders = {
        key: get_input('Enter the name of the {} folder'.format(key), key) for key in _folders
    }
    app = {}
    app['name'] = get_input('Enter the name of the app')
    app['desc'] = get_input('Enter the description of the app', '')
    app['prog'] = get_input('Enter the name of the program', 'main.py')

    # Prepare and create folders
    print(colored('\nCreating folders ...', 'cyan'))
    for key, folder in folders.items():
        with_init = True if key == 'experiments' or key == 'repositories' else False
        create_folder(ROOT, folder, with_init=with_init)

    # Create Entry Point
    print(colored('Creating main entry point ...', 'cyan'))
    attrs = {
        'klass': inflection.camelize(app['name'].replace(' ', '_')),
        'name': app['name'],
        'config_path': folders['config']
    }
    create_from_stub('App.stub', os.path.join(ROOT, app['prog']), attrs)

    # Creat config items
    print(colored('Creating config files ...', 'cyan'))
    app_cfg = {
        'prog': app['prog'],
        'description': app['desc'],
        'logging': {
            'level': 'info',
            'format': '[%(asctime)s] - [%(levelname)s] - [%(module)s] - %(message)s',
            'filename': '{}.log'.format(app['name'].replace(' ', '_').lower()),
            'path': folders['logs'],
            'max_bytes': 1024*1024,
            'backup_count': 10
        },
        'experiments': {'path': folders['experiments']}
    }
    storage_cfg = {
        'datastore': {
            'drivername': 'sqlite',
            'database': '{}.db'.format(app['name']),
            'username': None,
            'password': None,
            'host': None,
            'port': None
        },
        'migrations': {'path': folders['migrations']},
        'repositories': {'path': folders['repositories']}
    }

    create_config_file(ROOT, folders['config'], 'app.json', app_cfg)
    create_config_file(ROOT, folders['config'], 'storage.json', storage_cfg)

    # Create Migrations
    print(colored('Creating migrations ...', 'cyan'))
    _create_migration(
        folders['migrations'],
        'create_experiments',
        upgrade="""with self.schema.create('experiments') as table:
            table.increments('id')
            table.primary('id')
            table.string('name', 75)
            table.datetime('start')
            table.datetime('finished').nullable()
            table.string('config_file').nullable()
            table.text('config_content').nullable()""",
        down="self.schema.drop_if_exists('experiments')"
    )
    sleep(1)
    _create_migration(
        folders['migrations'],
        'create_testcase',
        upgrade=r"""with self.schema.create('testcases') as table:
            table.increments('id')
            table.primary('id')
            table.increments('experiment_id')
            table.foreign('experiment_id')\
                .references('id').on('experiments')\
                .on_delete('cascade')\
                .on_update('cascade')""",
        down="self.schema.drop_if_exists('testcases')"
    )
    sleep(1)
    _create_migration(
        folders['migrations'],
        'create_performance',
        upgrade=r"""with self.schema.create('performance') as table:
            table.big_increments('id')
            table.primary('id')
            table.string('label', 75)
            table.small_integer('level')
            table.string('type', 25)
            table.float('time')
            table.float('memory')
            table.float('peak_memory')
            table.integer('test_id')
            table.foreign('test_id')\
                .references('id').on('testcases')\
                .on_delete('cascade')\
                .on_update('cascade')""",
        down="self.schema.drop_if_exists('performance')"
    )

    # Create Repositories
    print(colored('Creating repositories ...', 'cyan'))
    _create_repository(
        folders['repositories'],
        'ExperimentRepository',
        'experiments',
        attributes=['name', 'config_file', 'start', 'config_content', 'finished'],
        nullable=['finished', 'config_content'],
        relationships={'tests': 'TestCaseRepository'}
    )
    _create_repository(
        folders['repositories'],
        'TestCaseRepository',
        'testcases',
        attributes=['iteration', 'experiment_id'],
        nullable=['experiment_id'],
        relationships={'performances': 'PerformanceRepository'}
    )
    _create_repository(
        folders['repositories'],
        'PerformanceRepository',
        'performance',
        attributes=['label', 'level', 'type', 'time', 'memory', 'peak_memory'],
    )

    # Done
    print(colored('Done.', 'yellow'))


if __name__ == '__main__':
    main()



"""Provides Config Holder and Config Loader modules to easily load/store config items.

Import :py:mod:`.Loader` and :py:class:`.Config`
so that they can be easly imported by other packages/modules.
"""
# flake8: noqa
from .Config import Config
from .Loader import Loader



"""Migration CLI commands to allow you to version control your database schema.

With Migrations your team is able to easily modify and share the
database schema to stay up to date. Migrations are typically paired
with the :py:mod:`.Schema` Builder which is inspired by the
`Laravel Schema Builder <https://laravel.com/docs/5.6/migrations#tables>`_
to easly manage your database's schema.


Generating Migrations
---------------------
Use the ``migration:make`` command  to create a new :py:class:`.Migration`. This will create the
following new :py:class:`.Migration` class in your ``migrations`` folder. In order to determine
the order of the migrations, each migration file name contains a timestamp.

Arguments:

====  ======================
name  Name of the migration
====  ======================


Migration Structure
-------------------
Each :py:class:`.Migration` class contains two methods: :py:meth:`.Migration.up` and
:py:meth:`.Migration.down`. In the :py:meth:`.Migration.up` method you add new tables,
columns, or indexes to your database. In the :py:meth:`.Migration.down` method you
should revert those changed made in the :py:meth:`.Migration.up` method.

Inside the :py:class:`.Migration` class you have access to the :py:mod:`.Schema` Builder
to easily create and modify tables. For more information check out its documentation:
:py:mod:`.Schema`

Running Migrations
------------------
To run the latest outstanding :py:class:`.Migration`, just use the ``migration:up`` command.
To revert the last :py:class:`.Migration` operation, just use the ``migration:down`` command.

To roll back all migrations and then execute all migrations, just use the ``migration:refresh``
command. This command effectively re-creates your entire database.

To see the status of the migrations, just use the ``migration:status`` command.
This would output something like this::

    |-------------------------------------------+--------|
    | Migration                                 | Ran?   |
    |-------------------------------------------+--------|
    | 20180814111005_create_users_table         | Yes    |
    | 20180815101334_add_avater_to_users_table  | No     |
    |-------------------------------------------+--------|
"""
from experimentum.Commands import command


@command(
    'Displays which migrations where migrated and which ones not.',
    help='Show the migration status.'
)
def status(app):
    """Show the status of the migrations.

    Arguments:
        app (App): Main Service Container
    """
    app.make('migrator').status()


@command(
    'Reset all migrations and then run them again to rebuild the data schema.',
    help='Rebuild the data schema (clears database).'
)
def refresh(app):
    """Refresh all migrations.

    Arguments:
        app (App): Main Service Container
    """
    app.make('migrator').refresh()


@command(
    'Upgrade the oldes migration which needs to be upgraded.',
    help='Upgrade oldest migration.'
)
def up(app):
    """Upgrade the oldes migration which needs to be upgraded.

    Arguments:
        app (App): Main Service Container
    """
    app.make('migrator').up()


@command(
    'Downgrade latest migration',
    help='Downgrade the latest migration which can be downgraded.'
)
def down(app):
    """Downgrade the latest migration which can be downgraded.

    Arguments:
        app (App): Main Service Container
    """
    app.make('migrator').down()


@command(
    'Create a new migration file in the migrations folder.',
    arguments={'name': {'help': 'Name of the migration'}},
    help='Make a new migration.'
)
def make(app, args):
    """Make a new migration file.

    Arguments:
        app (App): Main Service Container
        args (dict): Additional args
    """
    app.make('migrator').make(args.name.lower())
